@@ -1,4 +1,4 @@
-#!/usr/bin/python3.7
+#!/usr/bin/env python
 # -*-coding:utf8 -*
 
 """Module for UFPCA and MFPCA classes.
@@ -8,10 +8,14 @@ functional data. Univariate functional data and irregular functional data are
 concerned with UFPCA, whereas multivariate functional data with MFPCA.
 """
 import numpy as np
+import warnings
 
-from ...representation.functional_data import (DenseFunctionalData,
-                                               MultivariateFunctionalData)
-from ...misc.utils import integration_weights_
+from typing import Optional, List, Union
+
+from ...representation.functional_data import (
+    DenseFunctionalData, MultivariateFunctionalData
+)
+from ...misc.utils import _integration_weights
 
 from .fcp_tpa import FCPTPA
 
@@ -28,18 +32,26 @@ class UFPCA():
 
     Parameters
     ----------
-    n_components : int, float, None, default=None
+    n_components: int, float, None, default=None
         Number of components to keep.
-        if n_components if None, all components are kept::
+        If `n_components` is `None`, all components are kept::
             n_components == min(n_samples, n_features)
-        if n_components is int, n_components are kept.
-        if 0 < n_components < 1, select the number of components such that
+        If `n_components` is an integer, `n_components` are kept.
+        If `0 < n_components < 1`, select the number of components such that
         the amount of variance that needs to be explained is greater than
-        the percentage specified by n_components.
+        the percentage specified by `n_components`.
+    method: str, {'covariance', 'inner-product'}, default='covariance'
+        Method used to estimate the eigencomponents. If
+        ``method == 'covariance'``, the estimation is based on an
+        eigendecomposition of the covariance operator. If
+        ``method == 'inner-product'``, the estimation is based on an
+        eigendecomposition of the inner-product matrix.
+    normalize: bool, default=False
+        Perform a normalization of the data.
 
     Attributes
     ----------
-    eigenvalues: array, shape = (n_components, )
+    eigenvalues: array, shape = (n_components,)
         The singular values corresponding to each of selected components.
     eigenfunctions: DenseFunctionalData
         Principal axes in feature space, representing the directions of
@@ -49,28 +61,35 @@ class UFPCA():
     covariance: DenseFunctionalData
         An estimation of the covariance of the training data based on the
         results of the functional principal components analysis.
-
     """
 
-    def __init__(self, n_components=None):
+    def __init__(
+        self,
+        n_components: Union[int, float, None] = None,
+        method: str = 'covariance',
+        normalize: bool = False
+    ) -> None:
         """Initaliaze UFPCA object."""
         self.n_components = n_components
+        self.method = method
+        self.normalize = normalize
+        self.weights = 1
 
-    def fit(self, data, mean=None, covariance=None, **kwargs):
-        """Fit the model on data.
+    def fit(
+        self,
+        data: DenseFunctionalData,
+        **kwargs
+    ) -> None:
+        """Estimate the eigencomponents of the data.
 
         Parameters
         ----------
         data: DenseFunctionalData
-            Training data
-        mean: DenseFunctionalData, default=None
-            An estimation of the mean of the training data.
-        covariance: DenseFunctionalData, default=None
-            An estimation of the covariance of the training data.
+            Training data used to estimate the eigencomponents.
 
         Keyword Args
         ------------
-        method: str, default='LocalLinear'
+        method: str, default=None
             Smoothing method.
         kernel: str, default='gaussian'
             Kernel used for the smoothing.
@@ -89,17 +108,24 @@ class UFPCA():
             'degree': kwargs.get('degree', 2),
             'n_basis': kwargs.get('n_basis', 10)
         }
-        self._fit(data)
 
-    def _fit(self, data, mean=None, covariance=None):
-        """Dispatch to the right submethod depending on the input."""
-        if isinstance(data, DenseFunctionalData):
-            self._fit_uni(data, mean, covariance)
+        if not isinstance(data, DenseFunctionalData):
+            raise TypeError('UFPCA only support DenseFunctionalData object!')
+        if self.method == 'covariance':
+            self._fit_covariance(data)
+        elif self.method == 'inner-product':
+            self._fit_inner_product(data)
         else:
-            raise TypeError('UFPCA only support DenseFunctionalData'
-                            ' object!')
+            raise NotImplementedError(
+                f"{self.method} method not implemented."
+            )
 
-    def _fit_uni(self, data, mean=None, covariance=None):
+    def _fit_covariance(
+        self,
+        data: DenseFunctionalData,
+        mean: Optional[DenseFunctionalData] = None,
+        covariance: Optional[DenseFunctionalData] = None
+    ) -> None:
         """Univariate Functional PCA.
 
         Parameters
@@ -121,18 +147,26 @@ class UFPCA():
         TODO : Add possibility to smooth the eigenfunctions.
 
         """
+        if self.normalize:
+            data, weights = data.normalize(use_argvals_stand=True)
+            self.weights = weights
+
         smoothing_method = self.smoothing_parameters['method']
         if mean is None:
-            mean = data.mean(smooth=smoothing_method,
-                             **self.smoothing_parameters)
+            mean = data.mean(
+                smooth=smoothing_method,
+                **self.smoothing_parameters
+            )
         if covariance is None:
-            covariance = data.covariance(mean=mean,
-                                         smooth=smoothing_method,
-                                         **self.smoothing_parameters)
+            covariance = data.covariance(
+                mean=mean,
+                smooth=smoothing_method,
+                **self.smoothing_parameters
+            )
 
         # Choose the W_j's and the S_j's (Ramsey and Silverman, 2005)
         argvals = data.argvals['input_dim_0']
-        weight = integration_weights_(argvals, method='trapz')
+        weight = _integration_weights(argvals, method='trapz')
 
         # Compute the eigenvalues and eigenvectors of W^{1/2}VW^{1/2}
         weight_sqrt = np.diag(np.sqrt(weight))
@@ -154,8 +188,9 @@ class UFPCA():
 
         # Slice eigenvalues and compute eigenfunctions = W^{-1/2}U
         eigenvalues = eigenvalues[:npc]
-        eigenfunctions = np.transpose(np.dot(weight_invsqrt,
-                                             np.fliplr(eigenvectors)[:, :npc]))
+        eigenfunctions = np.transpose(
+            np.dot(weight_invsqrt, np.fliplr(eigenvectors)[:, :npc])
+        )
         # Compute estimation of the covariance
         temp = np.dot(np.transpose(eigenfunctions), np.diag(eigenvalues))
         cov = np.dot(temp, eigenfunctions)
@@ -168,7 +203,41 @@ class UFPCA():
         self.mean = mean
         self.covariance = DenseFunctionalData(new_argvals_2, cov[np.newaxis])
 
-    def transform(self, data, method='PACE'):
+    def _fit_inner_product(
+        self,
+        data: DenseFunctionalData
+    ) -> None:
+        """Univariate Functional PCA using inner-product matrix decomposition.
+
+        Parameters
+        ----------
+        data: DenseFunctionalData
+            Training data used to estimate the eigencomponents.
+
+        """
+        # Compute inner-product matrix
+
+        # Diagonalization of the inner-product matrix
+
+        # Estimation of the eigenvalues
+
+        # estimation of the eigenfunctions
+        pass
+
+    def _select_number_eigencomponents(self):
+        """Select the number of eigencomponents.
+        
+        Notes
+        -----
+        Put into utils?
+        """
+        pass
+
+    def transform(
+        self,
+        data: DenseFunctionalData,
+        method: str = 'PACE'
+    ) -> np.ndarray:
         """Apply dimensionality reduction to data.
 
         The functional principal components scores are given by:
@@ -201,9 +270,17 @@ class UFPCA():
 
         """
         # TODO: Add checkers
+        if self.normalize:
+            values = data.values / self.weights
+            data = DenseFunctionalData(data.argvals, values)
+
         data_unmean = data.values - self.mean.values
 
         if method == 'PACE':
+            warnings.warn((
+                "The implementation of PACE is not sure, prefer to use "
+                "`method=NumInt`."
+            ))
             sigma_inv = np.linalg.inv(
                 self.covariance.values[0] + data.var_noise * np.diagflat(
                     np.ones(shape=self.covariance.values[0].shape[0]))
@@ -219,7 +296,10 @@ class UFPCA():
 
         return scores
 
-    def inverse_transform(self, scores):
+    def inverse_transform(
+        self,
+        scores: np.ndarray
+    ) -> DenseFunctionalData:
         """Transform the data back to its original space.
 
         Return a DenseFunctionalData data_original whose transform would
@@ -262,6 +342,8 @@ class MFPCA():
         if 0 < n_components < 1, select the number of components such that
         the amount of variance that needs to be explained is greater than
         the percentage specified by n_components.
+    normalize: bool, default=False
+        Perform a normalization of the data.
 
     Attributes
     ----------
@@ -290,11 +372,20 @@ class MFPCA():
 
     """
 
-    def __init__(self, n_components=None):
+    def __init__(
+        self,
+        n_components: List[Union[int, float, None]] = None,
+        normalize: bool = False
+    ) -> None:
         """Initialize MFPCA object."""
         self.n_components = n_components
+        self.normalize = normalize
 
-    def fit(self, data, method='NumInt'):
+    def fit(
+        self,
+        data: MultivariateFunctionalData,
+        method: str = 'NumInt'
+    ) -> None:
         """Fit the model with X.
 
         Parameters
@@ -307,7 +398,11 @@ class MFPCA():
         """
         self._fit(data, method)
 
-    def _fit(self, data, method='NumInt'):
+    def _fit(
+        self,
+        data: MultivariateFunctionalData,
+        method: str = 'NumInt'
+    ) -> None:
         """Dispatch to the right submethod depending on the input."""
         # TODO: Different possiblity for n_components
         if isinstance(data, MultivariateFunctionalData):
@@ -316,7 +411,11 @@ class MFPCA():
             raise TypeError('MFPCA only support MultivariateFunctionalData'
                             ' object!')
 
-    def _fit_multi(self, data, method='NumInt'):
+    def _fit_multi(
+        self,
+        data: MultivariateFunctionalData,
+        method: str = 'NumInt'
+    ) -> None:
         """Multivariate Functional PCA.
 
         Parameters
@@ -337,7 +436,7 @@ class MFPCA():
         ufpca_list, scores = [], []
         for function, n in zip(data, self.n_components):
             if function.n_dim == 1:
-                ufpca = UFPCA(n_components=n)
+                ufpca = UFPCA(n_components=n, normalize=self.normalize)
                 ufpca.fit(data=function, method='GAM')
                 scores_uni = ufpca.transform(data=function, method='NumInt')
             elif function.n_dim == 2:
@@ -394,7 +493,11 @@ class MFPCA():
         self.eigenvectors = eigenvectors
         self.basis = MultivariateFunctionalData(basis_multi)
 
-    def transform(self, data, method='NumInt'):
+    def transform(
+        self,
+        data: MultivariateFunctionalData,
+        method: str = 'NumInt'
+    ) -> np.ndarray:
         """Apply dimensionality reduction to data.
 
         Parameters
@@ -416,7 +519,10 @@ class MFPCA():
         scores_uni = np.concatenate(scores_uni, axis=1)
         return np.dot(scores_uni, self.eigenvectors)
 
-    def inverse_transform(self, scores):
+    def inverse_transform(
+        self,
+        scores: np.ndarray
+    ) -> MultivariateFunctionalData:
         """Transform the data back to its original space.
 
         Return a MultivariateFunctionalData data_original whose transform would
